use std::fs::OpenOptions;
use clap::Values;
use std::io::{self, Write, BufRead, Seek, SeekFrom};
use crate::utils::{path, read};
use std::collections::HashMap;

pub struct ConfigSetArgs<'a> {
    pub name: Option<Values<'a>>,
    pub value: Option<Values<'a>>,
}

pub fn config_set(args: ConfigSetArgs) {
    // configure possible options and their associated category
    let mut option_categories: HashMap<&str, &str> = HashMap::new();
    option_categories.insert("force_insecure", "core");
    option_categories.insert("token", "core");

    let name = args.name.unwrap().next().unwrap();
    let value = args.value.unwrap().next().unwrap();

    // get category of option
    let category = option_categories.get(name);
    if category.is_none() {
        eprintln!("fatal: '{}' is not a valid option.", name);
        std::process::exit(1);
    }

    write_option_in_cat(category.unwrap(), name, value);
}


pub fn find_option_in_cat(category: &str, option: &str) -> Option<String> {
    let mut config = path::nextsync();
    config.push("config");

    let mut in_target_category = false;
    if let Ok(mut lines) = read::read_lines(config) {

        for line in lines {
            if let Ok(line) = line {
                let trimmed_line = line.trim();

                if trimmed_line.starts_with('[') && trimmed_line.ends_with(']') {
                    in_target_category = trimmed_line == format!("[{}]", category);
                } else if in_target_category {
                    let parts: Vec<&str> = trimmed_line.splitn(2, '=').collect();
                    if parts.len() == 2 && parts[0].trim() == option {
                        return Some(parts[1].trim().to_string());
                    }
                }

            }
        }
    }

    None
}

pub fn write_option_in_cat(category: &str, option: &str, value: &str) -> io::Result<()> {
    let mut config = path::nextsync();
    config.push("config");

    let mut file = OpenOptions::new()
        .read(true)
        .write(true)
        .create(true)
        .open(&config)?;

    let mut in_target_category = false;
    let mut option_found = false;

    // Go to the beginning of the file
    file.seek(SeekFrom::Start(0))?;

    // Create a temporary file to hold the modified content
    let mut tmp_file = tempfile::Builder::new()
        .prefix(".nextsyncconfig")
        .tempfile()?;

    let reader = io::BufReader::new(&file);
    for line in reader.lines() {
        let line = line?;
        let trimmed_line = line.trim();

        if trimmed_line.starts_with('[') && trimmed_line.ends_with(']') {
            // if we were already in target category we are now leaving it
            // add option only if not found before
            if in_target_category && !option_found {
                writeln!(&mut tmp_file, "\t{} = {}", option, value)?;
            } else if !in_target_category {
                in_target_category = trimmed_line == format!("[{}]", category);
            }
        }

        if in_target_category && !option_found && trimmed_line.starts_with(&format!("{} =", option)) {
            // Option already exists, update its value
            writeln!(&mut tmp_file, "\t{} = {}", option, value)?;
            option_found = true;
        } else {
            // Write the original line
            writeln!(&mut tmp_file, "{}", line)?;
        }
    }

    // add to last category
    if in_target_category && !option_found {
        writeln!(&mut tmp_file, "\t{} = {}", option, value)?;
    }

    // if the category didn't exist create it and add the option
    if !in_target_category {
        writeln!(&mut tmp_file, "[{}]", category)?;
        writeln!(&mut tmp_file, "\t{} = {}", option, value)?;
    }

    // Flush and sync the temporary file to ensure data is written to disk
    tmp_file.flush()?;

    // Go back to the beginning of the file
    tmp_file.seek(SeekFrom::Start(0))?;
    file.seek(SeekFrom::Start(0))?;

    // Copy the contents of the temporary file to the original file
    io::copy(&mut tmp_file, &mut file)?;

    Ok(())

}

pub fn add_remote(name: &str, url: &str) -> io::Result<()> {
    let mut config = path::config();

    // check if there is already a remote with this name
    if get_remote(name).is_some()
    {
        eprintln!("error: remote {} already exists.", name);
        std::process::exit(3);
    }

    let mut file = OpenOptions::new()
        .read(true)
        .write(true)
        .create(true)
        .append(true)
        .open(config)?;

    writeln!(file, "[remote \"{}\"]", name)?;
    writeln!(file, "\turl = {}", url)?;

    Ok(())
}

pub fn get_remote(name: &str) -> Option<String> {
    find_option_in_cat(&format!("remote \"{}\"", name), "url")
}

/// return a vector of remote found in config file (e.g: ("origin", "https://example.com"))
pub fn get_all_remote() -> Vec<(String, String)> {
    let config = path::config();

    let mut remotes: Vec<(String, String)> = vec![];

    let mut in_remote = false;
    let mut remote_name = String::new();
    if let Ok(lines) = read::read_lines(config) {

        for line in lines {
            if let Ok(line) = line {
                let trimmed_line = line.trim();

                if trimmed_line.starts_with("[remote ") {
                    in_remote = true;
                    remote_name = trimmed_line.strip_prefix("[remote \"").unwrap().strip_suffix("\"]").unwrap().to_string();
                }
                else if trimmed_line.starts_with('[')
                {
                    in_remote = false;
                }
                else if in_remote {
                    let parts: Vec<&str> = trimmed_line.splitn(2, '=').collect();
                    if parts.len() == 2 {
                        remotes.push((remote_name.to_string(), parts[1].trim().to_string()))
                    }
                }

            }
        }
    }
    remotes
}

pub fn get_core(name: &str) -> Option<String> {
    find_option_in_cat("core", name)
}

pub fn add_core(name: &str, value: &str) -> io::Result<()> {
    let mut config = path::nextsync();
    config.push("config");

    let mut file = OpenOptions::new()
        .read(true)
        .write(true)
        .create(true)
        .append(true)
        .open(config)?;

    writeln!(file, "[core]")?;
    writeln!(file, "\t{} = {}", name, value)?;

    Ok(())
}

pub fn get(var: &str) -> Option<String> {
    let mut config = path::config();

    if let Ok(lines) = read::read_lines(config) {
        for line in lines {
            if let Ok(l) = line {
                if l.starts_with(var) {
                    let (_, val) = l.split_once(" ").unwrap();
                    return Some(val.to_owned());
                }
            }
        } 
    }
    None
}


