use std::io::{self, Read};
use std::fs::{self, File};
use std::io::Write;
use std::fs::OpenOptions;
use std::path::PathBuf;
use std::time::SystemTime;
use crate::commands::status::{State};
use crate::utils::into::IntoPathBuf;
use crate::utils::{path, read};
use crate::store::head;
use crate::store::object::{update_dates, add_node, rm_node};

use crate::store::object::object::ObjMethods;
use crate::store::object::object::Obj;

const HASH_EMPTY: &str = "d41d8cd98f00b204e9800998ecf8427e";

pub struct Blob {
    pub obj: Obj,
    data: Vec<String>, // content of the ref file
    file_hash: Option<String>, // hash of the file's content
}

//pub struct Blob {
//    r_path: PathBuf, // relative path
//    a_path: PathBuf, // absolute path
//    hash: String, // hash of relative path
//    file_hash: Option<String>,
//    obj_p: PathBuf, // path of the object file
//    data: Vec<String>, // content of the blob
//}



impl Blob {
    pub fn new(obj: Obj) -> Self {
        Self { 
            obj,
            data: vec![],
            file_hash: None,
        }
    }
    pub fn from_path<S>(r_path: S) -> Blob where S: IntoPathBuf {
        let r_path = r_path.into();
        Self {
            obj: Obj::from_path(r_path),
            data: vec![],
            file_hash: None,
        }                
    }

    fn get_file_hash(&mut self) -> String {
        if self.file_hash.is_none() {
            let bytes = std::fs::read(self.get_file_path()).unwrap();
            let hash = md5::compute(&bytes);
            self.file_hash = Some(format!("{:x}", hash))
        }
        self.file_hash.clone().unwrap()
    }

    /// read line of blob to get all informations and store them in self.data
    pub fn read_data(&mut self) {
        if self.data.len() == 0 {
            if let Ok(mut file) = File::open(self.get_obj_path()) {
                let mut buffer = String::new();
                let _ = file.read_to_string(&mut buffer);
                let data = buffer.rsplit(' ').collect::<Vec<_>>();
                for e in data {
                    self.data.push(String::from(e));
                }
                self.data.reverse();

                // remove \n of last element
                if let Some(last) = self.data.last_mut() {
                    if last.ends_with("\n") {
                        last.pop();
                    }
                }
            }
        }
    }

    fn get_data_index(&mut self, index: usize) -> String {
        self.read_data();
        if self.data.len() >= index + 1 { 
            self.data[index].clone()
        } else {
            String::new()
        }
    }

    fn saved_filename(&mut self) -> String {
        self.get_data_index(0)
    }

    pub fn saved_remote_ts(&mut self) -> String {
        self.get_data_index(1)
    }

    fn saved_local_size(&mut self) -> String {
        self.get_data_index(2)
    }

    fn saved_local_ts(&mut self) -> u64 {
        match self.get_data_index(3).as_str() {
            "" => 0,
            str => str.parse::<u64>().unwrap()
        }
    }

    fn saved_hash(&mut self) -> String {
        self.get_data_index(4)
    }

    fn has_same_size(&mut self) -> bool {
        let metadata = match fs::metadata(self.get_file_path()) {
            Ok(m) => m,
            Err(_) => return true,
        };

        if self.saved_local_size() == String::new() { return true; }
        metadata.len().to_string() == self.saved_local_size()
    }

    fn is_newer(&mut self) -> bool {
        let metadata = match fs::metadata(self.get_file_path()) {
            Ok(m) => m,
            Err(_) => return true,
        };

        let secs = metadata
            .modified()
            .unwrap()
            .duration_since(SystemTime::UNIX_EPOCH)
            .unwrap()
            .as_secs();
        if self.saved_local_ts() == 0 { return true; }
        secs > self.saved_local_ts()
    }

    fn has_same_hash(&mut self) -> bool {
        if self.saved_hash() == String::new() { return false; }
        let file_hash = self.get_file_hash().clone();
        self.saved_hash() == file_hash
    }

    pub fn has_changes(&mut self) -> bool {
        !self.has_same_size() || (self.is_newer() && !self.has_same_hash())
    }

    pub fn get_all_identical_blobs(&mut self) -> Vec<String> {
        // an empty file is a new file not the copy of another empty file
        if self.get_file_hash() == HASH_EMPTY {
            return vec![];
        }

        let refs_p = self.get_obj_path();
        let mut blobs: Vec<String> = vec![];
        if let Ok(lines) = read::read_lines(refs_p) {
            for line in lines {
                if let Ok(l) = line {
                    blobs.push(l.clone());
                }
            } 
        }
        blobs
    }

    pub fn status(&mut self, path_from: &mut Option<PathBuf>) -> State {
        let has_obj_ref = self.get_obj_path().exists();
        let blob_exists = self.get_file_path().exists();

        if has_obj_ref && !blob_exists {
            State::Deleted
        } else if !has_obj_ref && blob_exists {
            let identical_blobs = self.get_all_identical_blobs();
            if identical_blobs.len() != 0 {
                let identical_blob = Blob::from_path(identical_blobs[0].clone()).get_local_obj();

                if identical_blob.state == State::Deleted {
                    *path_from = Some(identical_blob.path);
                    State::Moved
                } else if identical_blob.state == State::Default {
                    *path_from = Some(identical_blob.path);
                    State::Copied
                } else {
                    State::New
                }
            } else {
                State::New
            }
        } else if !has_obj_ref && !blob_exists {
            State::Default
        } else if self.has_changes() {
            State::Modified
        } else {
            State::Default
        }
    }

    fn create_blob_ref(&mut self, ts_remote: &str) -> io::Result<()> {
        let metadata = fs::metadata(self.get_file_path())?;
        let secs = metadata
            .modified()
            .unwrap()
            .duration_since(SystemTime::UNIX_EPOCH)
            .unwrap()
            .as_secs();

        // build line with all needed properties
        let content = format!("{} {} {} {} {}",
                              self.get_name(),
                              ts_remote,
                              metadata.len().to_string(),
                              secs.to_string(),
                              self.get_file_hash());

        // create parent dir if needed
        let mut obj_path = self.get_obj_path();
        obj_path.pop();
        if !obj_path.exists() {
           fs::create_dir_all(obj_path)?; 
        }

        // open ref file
        let mut file = OpenOptions::new()
            .create_new(true)
            .write(true)
            .open(self.get_obj_path())?;

        writeln!(file, "{}", content)?;

        Ok(())
    }

    fn get_file_ref(&mut self) -> PathBuf {
        let mut refs_p = path::refs();
        let file_hash = self.get_file_hash().clone();
        let (dir, res) = file_hash.split_at(2);

        refs_p.push(dir);
        if !refs_p.exists() {
           let _ = fs::create_dir_all(refs_p.clone()); 
        }
        refs_p.push(res);
        refs_p
    }

    // create a file in .nextsync/refs with the hash of this blob that
    // redirect to the relative path
    fn create_hash_ref(&mut self) -> io::Result<()> {
        // todo check if the file has been modified for moved and copy 
        let refs_p = self.get_file_ref();

        let mut file = OpenOptions::new()
            .create(true)
            .write(true)
            .open(refs_p)?;

        // todo deal with duplicate content
        writeln!(file, "{}", self.get_relative_file_path().to_str().unwrap())?;
        Ok(())
    }

    pub fn create(&mut self, ts_remote: &str, up_parent: bool) -> io::Result<()> {

        // add blob reference to parent
        self.add_ref_to_parent();

        if let Err(err) = self.create_blob_ref(ts_remote.clone()) {
            eprintln!("err: saving blob ref of {}: {}", self.get_relative_file_path().display(), err);
        }

        if let Err(err) = self.create_hash_ref() {
            eprintln!("err: saving hash ref of {}: {}", self.get_relative_file_path().display(), err);
        }

        // update date for all parent
        if up_parent {
            if let Err(err) = update_dates(self.get_relative_file_path(), ts_remote) {
                eprintln!("err: updating parent date of {}: {}", self.get_relative_file_path().display(), err);
            }
        }

        Ok(())
    }

    pub fn update(&mut self, ts_remote: &str) -> io::Result<()> {

//        // remove old hash ref
//        let mut refs_p = path::refs();
//        let binding = self.saved_hash();
//        let (dir, res) = binding.split_at(2);
//        refs_p.push(dir);
//        refs_p.push(res);
//        if let Err(err) = fs::remove_file(refs_p) {
//            eprintln!("err: removing hash ref of {}: {}", self.r_path.clone().display(), err);
//        }
//
//        // creating new hash ref
//        if let Err(err) = self.create_hash_ref() {
//            eprintln!("err: saving hash ref of {}: {}", self.r_path.clone().display(), err);
//        }
//
//        // updating content of blob's ref
//        let metadata = fs::metadata(self.a_path.clone())?;
//        let secs = metadata
//            .modified()
//            .unwrap()
//            .duration_since(SystemTime::UNIX_EPOCH)
//            .unwrap()
//            .as_secs();
//
//        let mut content = self.saved_filename();
//        content.push_str(" ");
//        content.push_str(ts_remote);
//        content.push_str(" ");
//        content.push_str(&metadata.len().to_string());
//        content.push_str(" ");
//        content.push_str(&secs.to_string());
//        content.push_str(" ");
//        content.push_str(&self.get_file_hash());
//
//        let mut file = OpenOptions::new()
//            .write(true)
//            .open(self.obj_p.clone())?;
//
//        writeln!(file, "{}", &content)?;
        Ok(())
    }
}
